#! /usr/bin/env python

## Copyright 2000-2001 by LivingLogic AG, Bayreuth, Germany.
## Copyright 2000-2001 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of Living Logic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVING LOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""
This module simplifies running Python stuff as cron jobs.

There will be no more than one sisyphus job of a certain name
running at every given time. When the job is already running
and a second is started, the second one will quit immediately
if the first one hasn't exceeded its maximum allowed lifetime
yet. If it has exceeded the allowed lifetime the first job
will be killed and the second will start running.

In addition to that, logging features are provided.

Functions:
execute -- execute one or more Job objects.

Classes:
LogFile -- Used by the Job class for logging.
Job -- derived your own sub classes from Job and implement 
  the execute method.
"""

__version__ = tuple(map(int,"$Revision: 1.31 $"[11:-2].split(".")))
# $Source: /data/cvsroot/LivingLogic/sisyphus/sisyphus.py,v $

import sys
import os
import os.path
import exceptions
import stat
import time
import traceback

def _formatTime(timestamp):
	return time.strftime("%d/%b/%Y %H:%M:%S", time.localtime(timestamp))

class LogFile:
	"""
	A log file. All lines written to the file will be prepended with a time stamp.
	"""
	def __init__(self, job, filename, mode="a"):
		"""
		__init__(job, filename, mode="a") -- construct a Log file (will be opened on the first write)
		
		job -- the job
		filename -- the filename (this may include "~/" or "~foo/" as os.path.expanduser()
		  will be called.)
		mode -- The mode for opening the file (should be "w" or "a")
		"""
		self.job = job
		self.filename = filename
		self.mode = mode
		self.file = None

	def log(self, text):
		"""
		self.log(text) -- write text to the log file
		"""
		now = time.time()
		lines = text.split("\n")
		if not len(lines[-1]):
			del lines[-1]
		lines = [ "[%s]=[t+%.2fsec] %s\n" % (_formatTime(now), now-self.job.startTime, line) for line in lines ]
		if self.file is None:
			self.file = open(os.path.expanduser(self.filename), self.mode, 1)
		self.file.write("".join(lines))

class Job:
	"""
	A Job object executes a task once.
	
	The job announces its presence (and its process id) in a file that
	is stored in the ~/run directory. Logs will be created in the
	~/log directory (This can be changes by deriving new subclasses).

	There are three log files:
	1. ~/log/<jobname>_progress.log
	   The progress of one job execution is logged here.
	   This log file will be truncated at the start of every execution.
	   Use the logProgress() method for writing to this log file.
	2. ~/log/<jobname>_loop.log
	   One log line may be appended to the log after every job
	   execution.
	   Use the logLoop() method for this
	3. ~/log/<jobname>_error.log
	   Here exceptions happening during the execution of a job
	   will be logged.
	   This is done via the logError method with can be used
	   for reporting additional errors.

	To use this class, derive your own class from it and overwrite
	the execute method.
	"""

	pidFilenameTemplate = "~/run/%s.pid"
	loopFilenameTemplate = "~/log/%s_loop.log"
	errorFilenameTemplate = "~/log/%s_error.log"
	progressFilenameTemplate = "~/log/%s_progress.log"

	def __init__(self, maxRuntime=0, name=None):
		"""
		__init__(maxRuntime=0, name=None) -- construct a Job object.
		
		maxRuntime -- the maximum allowed runtime in seconds for this job.
		name -- the name to be used for the log files. If None, the name 
		  of the class will be used.
		"""
		self.startTime = time.time()
		self.maxRuntime = maxRuntime
		self.name = name or self.__class__.__name__
		self.pidFileWritten = 0
		self.pidFilename = os.path.expanduser(self.pidFilenameTemplate % self.name)
		self.loopLogfile = LogFile(self, self.loopFilenameTemplate % self.name)
		self.errorLogfile = LogFile(self, self.errorFilenameTemplate % self.name)
		self.progressLogfile = LogFile(self, self.progressFilenameTemplate % self.name, "w")

	def __writePid(self):
		"""
		creates the file containing the pid of the current process
		"""
		file = open(self.pidFilename, "w")
		file.write(str(os.getpid()))
		file.close()
		self.pidFileWritten = 1

	def __killPid(self):
		"""
		deletes the pid file
		"""
		if self.pidFileWritten:
			os.remove(self.pidFilename)

	def logLoop(self, text):
		"""
		log the message text to the loop and progress log.
		(The call to logLoop() should be the last statement 
		in the execute method.)
		"""
		self.loopLogfile.log(text)
		self.progressLogfile.log(text)

	def logProgress(self, text):
		"""
		log the message text to the progress log
		"""
		self.progressLogfile.log(text)

	def logError(self, error):
		"""
		log the error to the error log.
		error may be a string or an exception object.
		"""
		if isinstance(error, exceptions.Exception):
			tb = "\n" + "".join(traceback.format_tb(sys.exc_info()[-1]))
			name = error.__class__.__name__
			tb += "%s: %s" % (name, error)
			self.errorLogfile.log(tb)
		else:
			self.errorLogfile.log(error)

	def execute(self):
		"""
		this method executes the job once.
		At the end of the job you should write one line to the loop log.
		Overwrite in subclasses.
		"""
		self.logLoop("done")

	def handleExecution(self):
		"""
		handles executing the job including handling of duplicate or hanging jobs.
		This is the method to be called from the outside world.
		"""
		try: # is there a pid file from a running job?
			file = open(self.pidFilename, "r")
			lastModified = os.fstat(file.fileno())[stat.ST_MTIME]
			pid = int(file.read())
		except IOError: # no pid file => the job has been finished without problems
			self.__writePid()
			self.logProgress("no previous job running; here we go")
		else:
			if self.startTime-lastModified > self.maxRuntime: # the job is to old, so it probably hangs => kill it
				try:
					os.kill(pid, 9)
				except OSError: # there was no process
					pass
				self.__writePid()
				self.logError("killed previous job running with pid %d (ran %.2f seconds; %d allowed); here we go" % (pid, self.startTime-lastModified, self.maxRuntime))
			else:
				self.logError("Job still running (for %.2f seconds; %d allowed; started on %s) with pid %d (according to %s)" % (self.startTime-lastModified, self.maxRuntime, _formatTime(lastModified), pid, self.pidFilename))
				return

		try:
			self.execute()
		except Exception, ex:
			self.logError(ex) # log the error
			self.logLoop("failed with %s" % ex.__class__.__name__) # loog the error to the lop loog too, because the job probably didn't have a chance to do it.
			if isinstance(ex, KeyboardInterrupt): # Really exit, if it was a CTRL-C
				self.__killPid()
				raise
		self.__killPid() # finished => remove the pid file

def execute(*jobs):
	"""
	sequentially execute several jobs.
	"""
	for job in jobs:
		job.handleExecution()

